#ifndef _DISPATCHER_SERVICE_H
#define _DISPATCHER_SERVICE_H
/*
 * FogLAMP dispatcher service.
 *
 * Copyright (c) 2021 Dianomic Systems
 *
 * Released under the Apache 2.0 Licence
 *
 * Author: Massimiliano Pinto
 */

#include <service_handler.h>
#include <management_client.h>
#include <management_api.h>
#include <reading.h>
#include <storage_client.h>
#include <dispatcher_api.h>
#include <controlrequest.h>
#include <mutex>
#include <condition_variable>
#include <queue>
#include <pipeline_manager.h>
#include <rapidjson/document.h>

#define SERVICE_NAME		"FogLAMP Dispatcher"
#define SERVICE_TYPE		"Dispatcher"
#define DEFAULT_WORKER_THREADS	2

/**
 * The DispatcherService class.
 *
 * The main class responsible for managing requests, handling the queues and interfacing
 * to the other FogLAMP services.
 */
class DispatcherService : public ServiceAuthHandler
{
	public:
		DispatcherService(const std::string& name, const std::string& token = "");
		~DispatcherService();
		bool 			start(std::string& coreAddress,
					      unsigned short corePort);
		void 			stop(bool removeFromCore=true);
		void			shutdown();
		void			restart();
		bool			isRunning() { return !m_stopping; };
		void			cleanupResources();
		void			configChange(const std::string&,
						     const std::string&);
		void			registerCategory(const std::string& categoryName);
		StorageClient*		getStorageClient() { return m_storage; };
		bool			queue(ControlRequest *request);
		void			worker();
		bool			sendToService(const std::string& service, 
						const std::string& url,
						const std::string& payload,
						const std::string& sourceName,
						const std::string& sourceType);
		void			configChildCreate(const std::string& parent_category,
							const std::string&,
							const std::string&) {};
		void			configChildDelete(const std::string& parent_category,
							const std::string&) {};
		void			setDryRun() { m_dryRun = true; };

		/**
		 * Return the pipeline manager for the service.
		 */
		ControlPipelineManager	*getPipelineManager() { return m_pipelineManager; };

		void			registerTable(const std::string& table);
		void			rowInsert(const std::string& table, const rapidjson::Document& doc);
		void			rowUpdate(const std::string& table, const rapidjson::Document& doc);
		void			rowDelete(const std::string& table, const rapidjson::Document& doc);

	private:
		ControlRequest		*getRequest();

	private:
		Logger*				m_logger;
		bool				m_shutdown;
		DispatcherApi*			m_api;
		ManagementApi*			m_managementApi;
		StorageClient*			m_storage;
		std::map<std::string, bool>
						m_registerCategories;
		unsigned long			m_worker_threads;
		const std::string       	m_token;
		std::queue<ControlRequest *>	m_requests;
		std::mutex			m_mutex;
		std::condition_variable		m_cv;
		bool				m_stopping;
		bool				m_enable;
		bool				m_dryRun;
		bool				m_restartRequest;
		bool				m_removeFromCore;
		ControlPipelineManager		*m_pipelineManager;
};
#endif
